#!/usr/bin/env python3
import serial
import time
import sys
import zlib
import serial.tools.list_ports
import os
from PIL import Image
import struct

# Configuration
CHUNK_SIZE = 512
BAUD_RATE = 115200
TIMEOUT = 5  # Increased timeout to match enhanced logic
MAX_RETRIES = 3
SERIAL_PORT = None
DEFAULT_STATIC_IMAGE = "path/to/default_static_image.jpg"  # Set your default static image path here
FRAME_MAGIC = 0xAA55CC33  # Magic frame for real-time image (fixed byte sequence)
RT_INIT_DELAY = 0.2       # Delay after START_RTIMAGE
RT_HEADER_DELAY = 0.1     # Delay after sending header
RT_RESPONSE_TIMEOUT = 5.0 # Timeout for waiting for device response

# Enhanced logging functions
def log_debug(message):
    print(f"[DEBUG] {message}")

def log_error(message):
    print(f"[ERROR] {message}")

def log_success(message):
    print(f"[SUCCESS] {message}")

# Functions for Image Upload
def send_line(ser, line):
    """Send a line with proper encoding and flushing"""
    try:
        data = (line + '\n').encode('utf-8')
        ser.write(data)
        ser.flush()
        log_debug(f"Sent: {line}")
        return True
    except Exception as e:
        log_error(f"Failed to send line '{line}': {e}")
        return False

def wait_for_response(ser: serial.Serial, expected, timeout=3.0):
    """Wait for expected response with improved error handling"""
    start = time.time()
    buffer = ""
    
    while time.time() - start < timeout:
        if ser.in_waiting:
            try:
                data = ser.read(ser.in_waiting)
                buffer += data.decode('utf-8', errors='ignore')
                
                while '\n' in buffer:
                    line, buffer = buffer.split('\n', 1)
                    line = line.strip()
                    
                    if line:
                        log_debug(f"Received: {line}")
                        if expected in line:
                            log_success(f"Expected response received: {line}")
                            return True
                        elif line in ["CHUNK_FAIL", "UPLOAD_CANCELED", "ERROR", "ABORT"]:
                            log_error(f"Device reported error: {line}")
                            return False
                            
            except Exception as e:
                log_error(f"Error reading response: {e}")
                
        time.sleep(0.01)  # Small delay to prevent CPU spinning
    
    log_error(f"Timeout waiting for response: {expected}")
    return False

def wait_for_ready(ser, timeout=10.0):
    """Wait for device READY signal with extended timeout"""
    log_debug("Waiting for device READY...")
    start = time.time()
    buffer = ""
    
    while time.time() - start < timeout:
        if ser.in_waiting:
            try:
                data = ser.read(ser.in_waiting)
                buffer += data.decode('utf-8', errors='ignore')
                
                while '\n' in buffer:
                    line, buffer = buffer.split('\n', 1)
                    line = line.strip()
                    
                    if line:
                        log_debug(f"Device response: {line}")
                        if line == "READY":
                            log_success("Device is ready for upload")
                            return True
                        elif "ERROR" in line or "FAIL" in line:
                            log_error(f"Device reported error: {line}")
                            return False
                            
            except Exception as e:
                log_error(f"Error reading ready signal: {e}")
                
        time.sleep(0.01)
    
    log_error("Device did not respond with READY")
    return False

def initialize_serial_connection(port):
    """Initialize serial connection with proper settings"""
    try:
        ser = serial.Serial(
            port=port,
            baudrate=BAUD_RATE,
            timeout=TIMEOUT,
            bytesize=serial.EIGHTBITS,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            write_timeout=TIMEOUT
        )
        
        try:
            ser.dtr = False
            log_debug("DTR set to False")
        except Exception as e:
            log_error(f"Could not set DTR: {e}")
        
        try:
            ser.rts = True
            time.sleep(0.1)
            ser.rts = False
            log_debug("RTS toggled")
        except Exception as e:
            log_error(f"Could not toggle RTS: {e}")
        
        time.sleep(2)
        ser.reset_input_buffer()
        ser.reset_output_buffer()
        
        log_success(f"Serial connection initialized on {port}")
        return ser
        
    except Exception as e:
        log_error(f"Failed to initialize serial connection: {e}")
        return None

def upload_image(path, is_realtime=False):
    """Enhanced image upload function"""
    if not os.path.exists(path):
        log_error(f"File not found: {path}")
        return False

    total_size = os.path.getsize(path)
    log_debug(f"Uploading '{path}' ({total_size} bytes)")

    ser = initialize_serial_connection(SERIAL_PORT)
    if not ser:
        return False

    try:
        with open(path, "rb") as f:
            command = f"REALTIME_IMAGE:{total_size}" if is_realtime else f"IMAGE_UPLOAD:{total_size}"
            if not send_line(ser, command):
                return False
                
            if not wait_for_ready(ser):
                log_error("Upload failed: device not ready")
                return False

            timestart = time.time()
            offset = 0
            
            while offset < total_size:
                chunk = f.read(CHUNK_SIZE)
                data_len = len(chunk)
                
                if data_len == 0:
                    break
                    
                crc32 = zlib.crc32(chunk) & 0xFFFFFFFF
                header = f"{offset},{data_len},{crc32:08X}"

                success = False
                for attempt in range(MAX_RETRIES):
                    log_debug(f"Sending chunk @ {offset} (attempt {attempt + 1})")
                    
                    if not send_line(ser, f"CHUNK:{header}"):
                        continue
                        
                    if not wait_for_response(ser, "CHUNK_START", timeout=5.0):
                        log_error(f"Device not ready for chunk: {header}")
                        continue
                    
                    try:
                        bytes_written = ser.write(chunk)
                        ser.flush()
                        
                        if bytes_written != data_len:
                            log_error(f"Incomplete write: {bytes_written}/{data_len} bytes")
                            continue
                            
                    except Exception as e:
                        log_error(f"Failed to write chunk data: {e}")
                        continue
                    
                    if wait_for_response(ser, "CHUNK_OK", timeout=5.0):
                        log_success(f"Chunk @ {offset} sent successfully ({data_len} bytes)")
                        success = True
                        break
                    else:
                        log_error(f"CRC mismatch for chunk @ {offset} (attempt {attempt + 1})")
                        time.sleep(0.5)

                if not success:
                    log_error(f"Failed to send chunk @ {offset} after {MAX_RETRIES} attempts")
                    send_line(ser, "UPLOAD_ABORT")
                    return False

                offset += data_len

            elapsed_time = time.time() - timestart
            log_success(f"Upload completed in {elapsed_time:.2f} seconds")
            return True

    except Exception as e:
        log_error(f"Upload error: {e}")
        return False
    finally:
        try:
            ser.close()
        except:
            pass

# Functions for Image Checks
def is_baseline_jpeg(image_path):
    if not os.path.exists(image_path):
        log_error(f"File not found: {image_path}")
        return False
    try:
        with Image.open(image_path) as img:
            if img.format != 'JPEG':
                return False
            return 'progressive' not in img.info or not img.info['progressive']
    except Exception as e:
        log_error(f"Error checking JPEG format for {image_path}: {e}")
        return False

def has_dimensions(image_path, expected_width, expected_height):
    if not os.path.exists(image_path):
        return False
    try:
        with Image.open(image_path) as img:
            return img.size == (expected_width, expected_height)
    except Exception as e:
        log_error(f"Error checking dimensions for {image_path}: {e}")
        return False

def is_file_size_less_than(file_path, max_kilobytes):
    if not os.path.exists(file_path):
        return False
    try:
        file_size_kb = os.path.getsize(file_path) / 1024.0
        return file_size_kb < max_kilobytes
    except Exception as e:
        log_error(f"Error checking file size for {file_path}: {e}")
        return False

def check_image_requirements(image_path):
    """Comprehensive image validation"""
    log_debug(f"Checking image requirements for: {image_path}")
    
    if not os.path.exists(image_path):
        log_error(f"File not found: {image_path}")
        return False

    try:
        with Image.open(image_path) as img:
            if img.format != 'JPEG':
                log_error(f"File is not a JPEG (format: {img.format})")
                return False

            expected_width, expected_height = 240, 320
            if img.size != (expected_width, expected_height):
                log_error(f"Dimensions {img.size} don't match required {expected_width}x{expected_height}")
                return False

    except Exception as e:
        log_error(f"Error opening image: {e}")
        return False

    if not is_baseline_jpeg(image_path):
        log_error("JPEG is not baseline format")
        return False

    max_kb = 55
    if not is_file_size_less_than(image_path, max_kb):
        try:
            file_size_kb = os.path.getsize(image_path) / 1024.0
            log_error(f"File size ({file_size_kb:.2f} KB) exceeds {max_kb} KB limit")
        except:
            log_error(f"File size exceeds {max_kb} KB limit")
        return False

    log_success("All image requirements met")
    return True

def send_realtime_image(port, image_path):
    """Enhanced real-time image upload with better error handling"""
    if not check_image_requirements(image_path):
        log_error("Image requirements not met")
        return False

    log_debug(f"Starting real-time image upload: {image_path}")
    
    try:
        with open(image_path, "rb") as f:
            image_data = f.read()
            total_size = len(image_data)
            log_debug(f"Image size: {total_size} bytes")

        ser = initialize_serial_connection(port)
        if not ser:
            return False

        try:
            log_debug("Sending START_RTIMAGE command")
            if not send_line(ser, "START_RTIMAGE"):
                return False
            
            time.sleep(RT_INIT_DELAY)

            header = struct.pack('<II', FRAME_MAGIC, total_size)
            log_debug(f"Sending header: magic={hex(FRAME_MAGIC)}, length={total_size}")
            
            try:
                ser.write(header)
                ser.flush()
            except Exception as e:
                log_error(f"Failed to send header: {e}")
                return False

            log_debug("Waiting for ready signal ('R')")
            ready_received = False
            start_time = time.time()
            
            while time.time() - start_time < RT_RESPONSE_TIMEOUT:
                if ser.in_waiting:
                    try:
                        data = ser.read(ser.in_waiting)
                        log_debug(f"Received data: {data}")
                        
                        if b'R' in data:
                            log_success("Ready signal ('R') received")
                            ready_received = True
                            break
                        elif b'E' in data or b"ERROR" in data:
                            log_error("Device reported error before sending data")
                            return False
                            
                    except Exception as e:
                        log_error(f"Error reading ready signal: {e}")
                        
                time.sleep(0.01)

            if not ready_received:
                log_error(f"Timeout waiting for ready signal after {RT_RESPONSE_TIMEOUT}s")
                return False

            time.sleep(RT_HEADER_DELAY)
            if ser.in_waiting:
                try:
                    response = ser.read(ser.in_waiting).decode('utf-8', errors='ignore')
                    if "FRAME_LOST" in response:
                        log_error("FRAME_LOST received, aborting")
                        send_line(ser, "ABORT")
                        return False
                except Exception as e:
                    log_debug(f"Error checking for FRAME_LOST: {e}")

            log_debug(f"Sending image data: {total_size} bytes")
            try:
                bytes_sent = ser.write(image_data)
                ser.flush()
                
                if bytes_sent != total_size:
                    log_error(f"Incomplete data send: {bytes_sent}/{total_size} bytes")
                    return False
                    
                log_debug(f"Successfully sent {bytes_sent} bytes")
                
            except Exception as e:
                log_error(f"Failed to send image data: {e}")
                return False

            log_debug("Waiting for acknowledgment")
            start_time = time.time()
            
            while time.time() - start_time < RT_RESPONSE_TIMEOUT:
                if ser.in_waiting:
                    try:
                        data = ser.read(ser.in_waiting)
                        log_debug(f"Acknowledgment data: {data}")
                        
                        if b'K' in data:
                            log_success("Image sent successfully ('K' received)")
                            return True
                        elif b'E' in data:
                            log_error("Device reported error ('E' received)")
                            return False
                            
                    except Exception as e:
                        log_error(f"Error reading acknowledgment: {e}")
                        
                time.sleep(0.01)
                
            log_error(f"Timeout waiting for acknowledgment after {RT_RESPONSE_TIMEOUT}s")
            return False

        finally:
            try:
                ser.close()
                log_debug("Serial connection closed")
            except Exception as e:
                log_error(f"Error closing serial connection: {e}")

    except Exception as e:
        log_error(f"Unexpected error during real-time image upload: {e}")
        return False

def send_serial_data(serial_port, data, timeout=1):
    port = serial_port
    baud_rate = BAUD_RATE
    max_retries = 3
    retry_delay = 1

    for attempt in range(max_retries):
        try:
            ser = serial.Serial(
                port=port,
                baudrate=baud_rate,
                timeout=timeout,
                bytesize=serial.EIGHTBITS,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                write_timeout=timeout
            )
            time.sleep(0.5)
            if not ser.is_open:
                log_error(f"Failed to open port {port}")
                continue
            log_success(f"Connected to {port} at {baud_rate} baud")
            if isinstance(data, str):
                data = data.encode('utf-8')
            ser.reset_input_buffer()
            ser.reset_output_buffer()
            bytes_sent = ser.write(data)
            log_debug(f"Sent {bytes_sent} bytes: {data}")
            ser.flush()
            time.sleep(0.1)
            if ser.in_waiting:
                response = ser.read(ser.in_waiting).decode('utf-8', errors='ignore').strip()
                log_debug(f"Received: {response}")
            ser.close()
            log_success("Connection closed successfully")
            return True
        except serial.SerialException as e:
            log_error(f"Attempt {attempt + 1}/{max_retries}: Serial error: {e}")
            if attempt < max_retries - 1:
                log_debug(f"Retrying in {retry_delay} seconds...")
                time.sleep(retry_delay)
            else:
                log_error("Failed to communicate with device after all retries.")
                return False
        except Exception as e:
            log_error(f"An unexpected error occurred: {e}")
            return False
        finally:
            try:
                if 'ser' in locals() and ser.is_open:
                    ser.close()
            except:
                pass

def send_idle(port):
    return send_serial_data(port, "IDLE\n")

def send_qr_payload(port, amount, message):
    formatted_data = f'QR**Rs. {amount}**{message}**I am qr data to be displayed in 2d qr form\n'
    return send_serial_data(port, formatted_data)

def send_text_display(port, title, subtitle, message):
    formatted_data = f'TEXT**{title}**{subtitle}**{message}\n'
    return send_serial_data(port, formatted_data)

def send_loading(port, amount, message):
    formatted_data = f'WAIT**Rs. {amount}**{message}\n'
    return send_serial_data(port, formatted_data)

def send_success(port, title, message):
    formatted_data = f'PASS**{title}**{message}\n'
    return send_serial_data(port, formatted_data)

def send_failure(port, amount, message):
    formatted_data = f'FAIL**Rs. {amount}**{message}\n'
    return send_serial_data(port, formatted_data)

def send_warning(port, title, message):
    formatted_data = f'WARN**{title}**{message}\n'
    return send_serial_data(port, formatted_data)

def send_information(port, title, message):
    formatted_data = f'INFO**{title}**{message}\n'
    return send_serial_data(port, formatted_data)

def send_reset(port):
    try:
        send_serial_data(port, "RESET\n")
        time.sleep(1)
        send_serial_data(port, "CLEAR\n")
        time.sleep(0.5)
        return True
    except Exception as e:
        log_error(f"Error during reset: {e}")
        return False

def send_format(port):
    try:
        send_serial_data(port, "FORMAT\n")
        time.sleep(1)
        send_serial_data(port, "CLEAR\n")
        time.sleep(0.5)
        return True
    except Exception as e:
        log_error(f"Error during format: {e}")
        return False

def send_wake(port):
    try:
        return send_serial_data(port, "WAKE\n")
    except Exception as e:
        log_error(f"Error during wake command: {e}")
        return False

def send_screentimer(port, seconds):
    try:
        if not isinstance(seconds, int) or seconds < 30 or seconds > 300:
            log_error("Screen timer must be between 30 and 300 seconds")
            return False
        formatted_data = f'SCREENTIME**{seconds}\n'
        if not send_serial_data(port, formatted_data):
            return False
        return send_wake(port)
    except Exception as e:
        log_error(f"Error during screen timer command: {e}")
        return False

def send_static_image(port, image_path):
    try:
        if not check_image_requirements(image_path):
            log_error(f"Image requirements not met for {image_path}")
            return False
        global SERIAL_PORT
        SERIAL_PORT = port
        if not upload_image(image_path, is_realtime=False):
            log_error("Failed to upload static image")
            return False
        log_success(f"Static image {image_path} uploaded and displayed successfully")
        return True
    except Exception as e:
        log_error(f"Error during static image upload: {e}")
        return False

def send_clear_display(port):
    try:
        if not send_serial_data(port, "CLEAR\n"):
            log_error("Failed to send clear command")
            return False
        time.sleep(0.5)
        if os.path.exists(DEFAULT_STATIC_IMAGE):
            if not send_static_image(port, DEFAULT_STATIC_IMAGE):
                log_error("Failed to restore static image after clear")
                return False
        else:
            log_error(f"Default static image {DEFAULT_STATIC_IMAGE} not found")
        log_success("Real-time display cleared and static image restored")
        return True
    except Exception as e:
        log_error(f"Error during display clear: {e}")
        return False

def check_arg_n_run_cmd(port):
    if len(sys.argv) < 2:
        print("Usage: python phythonfile.py <command> [arguments]")
        print("\nAvailable commands:")
        print("  idle                    - Display idle screen")
        print("  qr <amount> <message>   - Send QR code")
        print("  static <path>           - Upload static image to device memory")
        print("  image <path>            - Upload image")
        print("  realtime <path>         - Upload and display image in real-time mode")
        print("  clear                   - Clear real-time display (restores static image)")
        print("  text <title> <subtitle> <message> - Display text")
        print("  loading <amount> <message> - Show loading screen")
        print("  success <title> <message> - Show success message")
        print("  failure <amount> <message> - Show failure message")
        print("  warning <title> <message> - Show warning message")
        print("  info <title> <message>    - Show information message")
        print("  reset                   - Reset device")
        print("  format                  - Format device")
        print("  wake                    - Wake up the device")
        print("  screentimer <seconds>   - Set screen timer (30-300 seconds)")
        return False

    command = sys.argv[1].lower()
    result = False

    static_image_available = os.path.exists(DEFAULT_STATIC_IMAGE)
    if static_image_available and not check_image_requirements(DEFAULT_STATIC_IMAGE):
        log_error(f"Default static image {DEFAULT_STATIC_IMAGE} does not meet requirements")
        static_image_available = False

    if command == "idle":
        result = send_idle(port)
    elif command == "qr":
        if len(sys.argv) != 4:
            print("Usage: python phythonfile.py qr <amount> <message>")
            return False
        result = send_qr_payload(port, sys.argv[2], sys.argv[3])
    elif command == "static":
        if len(sys.argv) != 3:
            print("Usage: python phythonfile.py static <path>")
            return False
        result = send_static_image(port, sys.argv[2])
    elif command == "image":
        if len(sys.argv) != 3:
            print("Usage: python phythonfile.py image <path>")
            return False
        if not check_image_requirements(sys.argv[2]):
            return False
        global SERIAL_PORT
        SERIAL_PORT = port
        result = upload_image(sys.argv[2])
    elif command == "realtime":
        if len(sys.argv) != 3:
            print("Usage: python phythonfile.py realtime <path>")
            return False
        result = send_realtime_image(port, sys.argv[2])
    elif command == "clear":
        result = send_clear_display(port)
    elif command == "text":
        if len(sys.argv) != 5:
            print("Usage: python phythonfile.py text <title> <subtitle> <message>")
            return False
        result = send_text_display(port, sys.argv[2], sys.argv[3], sys.argv[4])
    elif command == "loading":
        if len(sys.argv) != 4:
            print("Usage: python phythonfile.py loading <amount> <message>")
            return False
        result = send_loading(port, sys.argv[2], sys.argv[3])
    elif command == "success":
        if len(sys.argv) != 4:
            print("Usage: python phythonfile.py success <title> <message>")
            return False
        result = send_success(port, sys.argv[2], sys.argv[3])
    elif command == "failure":
        if len(sys.argv) != 4:
            print("Usage: python phythonfile.py failure <amount> <message>")
            return False
        result = send_failure(port, sys.argv[2], sys.argv[3])
    elif command == "warning":
        if len(sys.argv) != 4:
            print("Usage: python phythonfile.py warning <title> <message>")
            return False
        result = send_warning(port, sys.argv[2], sys.argv[3])
    elif command == "info":
        if len(sys.argv) != 4:
            print("Usage: python phythonfile.py info <title> <message>")
            return False
        result = send_information(port, sys.argv[2], sys.argv[3])
    elif command == "reset":
        result = send_reset(port)
    elif command == "format":
        result = send_format(port)
    elif command == "wake":
        result = send_wake(port)
    elif command == "screentimer":
        if len(sys.argv) != 3:
            print("Usage: python phythonfile.py screentimer <seconds>")
            return False
        try:
            seconds = int(sys.argv[2])
            result = send_screentimer(port, seconds)
        except ValueError:
            log_error("Screen timer must be a number")
            return False
    else:
        print(f"Unknown command: {command}")
        return False

    if result and command != "realtime" and static_image_available:
        if not send_static_image(port, DEFAULT_STATIC_IMAGE):
            log_error("Failed to display default static image after command")
            return False
    return result

def find_all_ch341_devices():
    ch341_vid = 0x1a86
    ch341_pids = [0x7523, 0x5523]
    found_devices = []
    log_debug("Searching for CH341 devices...")
    ports = serial.tools.list_ports.comports()
    for port in ports:
        if (port.vid is not None and port.pid is not None and 
            port.vid == ch341_vid and port.pid in ch341_pids):
            device_info = {
                "port": port.device,
                "description": port.description,
                "hwid": port.hwid,
                "vid": hex(port.vid),
                "pid": hex(port.pid)
            }
            found_devices.append(device_info)
    return found_devices

if __name__ == "__main__":
    print("------------ USE GUIDE -------------")
    print(f"Default static image: {DEFAULT_STATIC_IMAGE}")
    print("Available commands:")
    print("1. idle                    - Display idle screen")
    print("2. qr <amount> <message>   - Send QR code")
    print("3. static <path>           - Upload static image to device memory")
    print("4. image <path>            - Upload image")
    print("5. realtime <path>         - Upload and display image in real-time mode")
    print("6. clear                   - Clear real-time display (restores static image)")
    print("7. text <title> <subtitle> <message> - Display text")
    print("8. loading <amount> <message> - Show loading screen")
    print("9. success <title> <message> - Show success message")
    print("10. failure <amount> <message> - Show failure message")
    print("11. warning <title> <message> - Show warning message")
    print("12. info <title> <message>    - Show information message")
    print("13. reset                  - Reset device")
    print("14. format                 - Format device")
    print("15. wake                   - Wake up the device")
    print("16. screentimer <seconds>  - Set screen timer (30-300 seconds)")
    print("\nExamples:")
    print("  python phythonfile.py idle")
    print("  python phythonfile.py qr 123.23 'SCAN TO PAY'")
    print("  python phythonfile.py static 'path/to/image.jpg'")
    print("  python phythonfile.py image 'path/to/image.jpg'")
    print("  python phythonfile.py realtime 'path/to/image.jpg'")
    print("  python phythonfile.py clear")
    print("  python phythonfile.py text 'Main Title' 'Subtitle' 'Message'")
    print("  python phythonfile.py loading 560.50 'Please wait...'")
    print("  python phythonfile.py success 'SUCCESS!' 'Payment successful'")
    print("  python phythonfile.py failure 560.50 'Payment Failed'")
    print("  python phythonfile.py warning 'Device Not Ready' 'Please wait'")
    print("  python phythonfile.py info 'Important' 'Keep device connected'")
    print("  python phythonfile.py reset")
    print("  python phythonfile.py format")
    print("  python phythonfile.py wake")
    print("  python phythonfile.py screentimer 60")
    print("----------------- END OF GUIDE ----------------\n\n")

    try:
        ch341_devices = find_all_ch341_devices()
        no_of_ch341_devices = len(ch341_devices)

        if no_of_ch341_devices == 1:
            device = ch341_devices[0]
            print(f"Found CH341 device:")
            print(f"  Port: {device['port']}")
            print(f"  Description: {device['description']}")
            print(f"  Hardware ID: {device['hwid']}")
            print(f"  VID: {device['vid']}")
            print(f"  PID: {device['pid']}\n")
            
            SERIAL_PORT = device['port']
        elif no_of_ch341_devices > 1:
            print(f"Found {no_of_ch341_devices} CH341 device(s):")
            for i, device in enumerate(ch341_devices, 1):
                print(f"\nDevice {i}:")
                print(f"  Port: {device['port']}")
                print(f"  Description: {device['description']}")
                print(f"  Hardware ID: {device['hwid']}")
            while True:
                try:
                    selection = int(input(f"\nEnter the device number to connect (1-{no_of_ch341_devices}): "))
                    if 1 <= selection <= no_of_ch341_devices:
                        SERIAL_PORT = ch341_devices[selection - 1]['port']
                        break
                    else:
                        print("Invalid selection. Please try again.")
                except ValueError:
                    print("Invalid input. Please enter a number.")
        else:
            print("No CH341 devices found. Please check:")
            print("1. Is the device connected?")
            print("2. Are the drivers installed?")
            print("3. Is the device powered on?")
            sys.exit(1)
    
        if check_arg_n_run_cmd(SERIAL_PORT):
            log_success("Command executed successfully")
        else:
            log_error("Command execution failed")
                
    except Exception as e:
        log_error(f"An error occurred: {e}")
        sys.exit(1)